#!/usr/bin/env python
"""
A page that knows how to format DB queries
"""




import os
import logging
import threading

from WMCore.WebTools.Page import TemplatedPage
from WMCore.Database.DBFormatter import DBFormatter
from WMCore.Database.DBFactory import DBFactory
from WMCore.Database.Transaction import Transaction
from WMCore.WebTools.ConfigDBMap import ConfigDBMap

class DatabasePage(TemplatedPage, DBFormatter):
    """
    A page that knows how to format DB queries
    """
    def __init__(self, config = {}):
        """
        __DatabasePage__

        A page with a database connection (a WMCore.Database.DBFormatter) held
        in self.dbi. Look at the DBFormatter class for other handy helper
        methods, such as getBinds and formatDict.

        The DBFormatter class was originally intended to be extensively
        sub-classed, such that it's subclasses followed the DAO pattern. For web
        tools we do not generally do this, and you will normally access the
        database interface directly:

        binds = {'id': 123}
        sql = "select * from table where id = :id"
        result = self.dbi.processData(sql, binds)
        return self.formatDict(result)

        Although following the DAO pattern is still possible and encouraged
        where appropriate. However, if you want to use the DAO pattern it may be
        better to *not* expose the DAO classes and have a normal DatabasePage
        exposed that passes the database connection to all the DAO's.
        """
        TemplatedPage.__init__(self, config)
        dbConfig = ConfigDBMap(config)
        conn = DBFactory(self, dbConfig.getDBUrl(), dbConfig.getOption()).connect()
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.INFO)
        DBFormatter.__init__(self, logger, conn)
        myThread = threading.currentThread()
        myThread.transaction = Transaction(conn)
        myThread.transaction.commit()
        return



#!/usr/bin/env python
"""
_SetStateTime_

Oracle implementation of Jobs.SetStateTime
"""




from WMCore.WMBS.MySQL.Jobs.SetStateTime import SetStateTime as MySQLSetStateTime

class SetStateTime(MySQLSetStateTime):
    pass



#/usr/bin/env python2.4
"""
_Destroy_

"""




from WMCore.Agent.Database.DestroyAgentBase import DestroyAgentBase
from WMCore.Agent.Database.Oracle.Create import Create

class Destroy(DestroyAgentBase):
    def __init__(self, logger = None, dbi = None):
        """
        _init_

        Call the base class's constructor and add all necessary tables for
        deletion,
        """
        DestroyAgentBase.__init__(self, logger, dbi)

        for tableName in Create.sequenceTables:
            seqname = '%s_SEQ' % tableName
            self.create["%s%s" % (Create.seqStartNum, seqname)] = \
                           "DROP SEQUENCE %s"  % seqname
            # triggers have to be deleted first
            triggerName = '%s_TRG' % tableName
            self.create["%s%s" % ('00', triggerName)] = \
                           "DROP TRIGGER %s"  % triggerName



"""
_GetPNNtoPSNMapping_

Oracle implementation of Locations.GetPNNtoPSNMapping

"""
from __future__ import (print_function, division)

from WMCore.WMBS.MySQL.Locations.GetPNNtoPSNMapping import GetPNNtoPSNMapping as MySQLGetPNNtoPSNMapping

class GetPNNtoPSNMapping(MySQLGetPNNtoPSNMapping):
    """
    Same as MySQL version

    """
    pass



#!/usr/bin/env python
"""
_CompleteFiles_

MySQL implementation of Subscription.CompleteFiles
"""

from WMCore.Database.DBFormatter import DBFormatter

class CompleteFiles(DBFormatter):
    sql = """INSERT IGNORE INTO wmbs_sub_files_complete (subscription, fileid)
               VALUES (:subscription, :fileid)"""

    delAcq = """DELETE FROM wmbs_sub_files_acquired
                WHERE subscription = :subscription AND
                      fileid = :fileid"""

    delAva = """DELETE FROM wmbs_sub_files_available
                WHERE subscription = :subscription AND
                      fileid = :fileid"""

    delFai = """DELETE FROM wmbs_sub_files_failed
                WHERE subscription = :subscription AND
                      fileid = :fileid"""

    def execute(self, subscription = None, file = None, conn = None,
                transaction = False):
        binds = self.getBinds(subscription = subscription, fileid = file)
        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)
        self.dbi.processData(self.delAcq, binds, conn = conn,
                             transaction = transaction)
        self.dbi.processData(self.delAva, binds, conn = conn,
                             transaction = transaction)
        self.dbi.processData(self.delFai, binds, conn = conn,
                             transaction = transaction)

        return



"""
File       : Pipeline.py
Description: Provides 2 basic classes:
             - Functor:  A class to create function calls from a function object
                         and arbitrary number of arguments
             - Pipeline: A class to provide building blocks for creating functional
                         pipelines for cumulative execution on an arbitrary object
"""

# futures
from __future__ import division, print_function

from builtins import object
from functools import reduce


class Functor(object):
    """
    A simple functor class used to construct a function call which later to be
    applied on an (any type) object.
    NOTE:
        It expects a function in the constructor and an (any type) object
        passed to the run or __call__ methods, which methods once called they
        construct and return the following function:
        func(obj, *args, **kwargs)
    NOTE:
        All the additional arguments which the function may take must be set in
        the __init__ method. If any of them are passed during run time an error
        will be raised.

    :func:
        The function to which the rest of the constructor arguments are about
        to be attached and then the newly created function will be returned.
        - The function needs to take at least one parameter since the object
        passed to the run/__call__ methods will always be put as a first
        argument to the function.

    :Example:

    def adder(a, b, *args, **kwargs):
        if args:
            print("adder args: %s" % args)
        if kwargs:
            print("adder kwargs: %s" % kwargs)
        res = a + b
        return res

    >>> x=Functor(adder, 8, 'foo', bar=True)
    >>> x(2)
    adder args: foo
    adder kwargs: {'bar': True}
    adder res: 10
    10

    >>> x
    <Pipeline.Functor instance at 0x7f319bbaeea8>


    """
    def __init__(self, func, *args, **kwargs):
        """
        The init method for class Functor
        """
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __call__(self, obj):
        """
        The call method for class Functor
        """
        return self.run(obj)

    def run(self, obj):
        return self.func(obj, *self.args, **self.kwargs)


class Pipeline(object):
    """
    A simple Functional Pipeline Class: applies a set of functions to an object,
    where the output of every previous function is an input to the next one.
    """
    # NOTE:
    #    Similar and inspiring approaches but yet some different implementations
    #    are discussed in the following two links [1] & [2]. With a quite good
    #    explanation in [1], which helped a lot. All in all at the bottom always
    #    sits the reduce function.
    #    [1]
    #    https://softwarejourneyman.com/python-function-pipelines.html
    #    [2]
    #    https://gitlab.com/mc706/functional-pipeline

    def __init__(self, funcLine=None, name=None):
        """
        :funcLine: A list of functions or Functors of function + arguments (see
                   the Class definition above) that are to be applied sequentially
                   to the object.
                   - If any of the elements of 'funcLine' is a function, a direct
                   function call with the object as an argument is performed.
                   - If any of the elements of 'funcLine' is a Functor, then the
                   first argument of the Functor constructor is the function to
                   be evaluated and the object is passed as a first argument to
                   the function with all the rest of the arguments passed right
                   after it eg. the following Functor in the funcLine:

                   Functor(func, 'foo', bar=True)

                   will result in the following function call later when the
                   pipeline is executed:

                   func(obj, 'foo', bar=True)

        :Example:
            (using the adder function from above and an object of type int)

        >>> pipe = Pipeline([Functor(adder, 5),
                             Functor(adder, 6),
                             Functor(adder, 7, "extraArg"),
                             Functor(adder, 8, update=True)])

        >>> pipe.run(1)
        adder res: 6
        adder res: 12
        adder args: extraArg
        adder res: 19
        adder kwargs: {'update': True}
        adder res: 27
        """
        self.funcLine = funcLine or []
        self.name = name

    def getPipelineName(self):
        """
        __getPipelineName__
        """
        name = self.name or "Unnamed Pipeline"
        return name

    def run(self, obj):
        return reduce(lambda obj, functor: functor(obj), self.funcLine, obj)



#!/usr/bin/env python
"""
WorkQueue splitting by block

"""
from __future__ import print_function, division

import logging
from math import ceil

from WMCore.WorkQueue.Policy.Start.StartPolicyInterface import StartPolicyInterface
from WMCore.WorkQueue.WorkQueueExceptions import WorkQueueWMSpecError
from WMCore.WorkQueue.WorkQueueUtils import makeLocationsList
from WMCore import Lexicon


class Block(StartPolicyInterface):
    """Split elements into blocks"""

    def __init__(self, **args):
        StartPolicyInterface.__init__(self, **args)
        self.args.setdefault('SliceType', 'NumberOfFiles')
        self.args.setdefault('SliceSize', 1)
        self.lumiType = "NumberOfLumis"

        # Initialize a list of sites where the data is
        self.sites = []

        # Initialize modifiers of the policy
        self.blockBlackListModifier = []

    def split(self):
        """Apply policy to spec"""
        dbs = self.dbs()
        for block in self.validBlocks(self.initialTask, dbs):
            # set the parent flag for processing only for clarity on the couch doc
            parentList = {}
            parentFlag = False
            # TODO this is slow process needs to change in DBS3
            if self.initialTask.parentProcessingFlag():
                parentFlag = True
                parentBlocks = dbs.listBlockParents(block["block"])
                for blockName in parentBlocks:
                    if self.initialTask.getTrustSitelists().get('trustlists'):
                        parentList[blockName] = self.sites
                    else:
                        blockLocations = self.blockLocationRucioPhedex(blockName)
                        parentList[blockName] = self.cric.PNNstoPSNs(blockLocations)

            # there could be 0 event files in that case we can't estimate the number of jobs created.
            # We set Jobs to 1 for that case.
            # If we need more realistic estimate, we need to dry run the spliting the jobs.
            estimateJobs = max(1, ceil(block[self.args['SliceType']] / self.args['SliceSize']))

            self.newQueueElement(Inputs={block['block']: self.data.get(block['block'], [])},
                                 ParentFlag=parentFlag,
                                 ParentData=parentList,
                                 NumberOfLumis=int(block[self.lumiType]),
                                 NumberOfFiles=int(block['NumberOfFiles']),
                                 NumberOfEvents=int(block['NumberOfEvents']),
                                 Jobs=estimateJobs,
                                 OpenForNewData=False,
                                 NoInputUpdate=self.initialTask.getTrustSitelists().get('trustlists'),
                                 NoPileupUpdate=self.initialTask.getTrustSitelists().get('trustPUlists')
                                )

    def validate(self):
        """Check args and spec work with block splitting"""
        StartPolicyInterface.validateCommon(self)

        if not self.initialTask.inputDataset():
            raise WorkQueueWMSpecError(self.wmspec, 'No input dataset')

    def validBlocks(self, task, dbs):
        """Return blocks that pass the input data restriction"""
        datasetPath = task.getInputDatasetPath()
        validBlocks = []

        blockWhiteList = task.inputBlockWhitelist()
        blockBlackList = task.inputBlockBlacklist()
        runWhiteList = task.inputRunWhitelist()
        runBlackList = task.inputRunBlacklist()
        if task.getLumiMask():  # if we have a lumi mask get only the relevant blocks
            maskedBlocks = self.getMaskedBlocks(task, dbs, datasetPath)
        if task.getTrustSitelists().get('trustlists'):
            siteWhitelist = task.siteWhitelist()
            siteBlacklist = task.siteBlacklist()
            self.sites = makeLocationsList(siteWhitelist, siteBlacklist)

        blocks = []
        # Take data inputs or from spec
        if not self.data:
            if blockWhiteList:
                self.data = dict((block, []) for block in blockWhiteList)
            else:
                self.data = {datasetPath: []}  # same structure as in WorkQueueElement

        for data in self.data:
            if data.find('#') > -1:
                Lexicon.block(data)  # check block name
                datasetPath = str(data.split('#')[0])
                blocks.append(str(data))
            else:
                Lexicon.dataset(data)  # check dataset name
                for block in dbs.listFileBlocks(data):
                    blocks.append(str(block))

        for blockName in blocks:
            # check block restrictions
            if blockWhiteList and blockName not in blockWhiteList:
                continue
            if blockName in blockBlackList:
                continue
            if blockName in self.blockBlackListModifier:
                # Don't duplicate blocks rejected before or blocks that were included and therefore are now in the blacklist
                continue
            if task.getLumiMask() and blockName not in maskedBlocks:
                logging.warning("Block %s doesn't pass the lumi mask constraints", blockName)
                self.rejectedWork.append(blockName)
                continue

            block = self._getBlockSummary(dbs, datasetPath, blockName)
            if not block:
                continue

            # check lumi restrictions
            if task.getLumiMask():
                accepted_lumis = sum([len(maskedBlocks[blockName][lfn].getLumis()) for lfn in maskedBlocks[blockName]])
                # use the information given from getMaskedBlocks to compute che size of the block
                block['NumberOfFiles'] = len(maskedBlocks[blockName])
                # ratio =  lumis which are ok in the block / total num lumis
                ratioAccepted = accepted_lumis / block['NumberOfLumis']
                block['NumberOfEvents'] = block['NumberOfEvents'] * ratioAccepted
                block[self.lumiType] = accepted_lumis
            # check run restrictions
            elif runWhiteList or runBlackList:
                # listRunLumis returns a dictionary with the lumi sections per run
                runLumis = dbs.listRunLumis(block=block['block'])
                runs = set(runLumis.keys())
                recalculateLumiCounts = False
                if len(runs) > 1:
                    # If more than one run in the block
                    # Then we must calculate the lumi counts after filtering the run list
                    # This has to be done rarely and requires calling DBS file information
                    recalculateLumiCounts = True

                # apply blacklist
                runs = runs.difference(runBlackList)
                # if whitelist only accept listed runs
                if runWhiteList:
                    runs = runs.intersection(runWhiteList)
                # any runs left are ones we will run on, if none ignore block
                if not runs:
                    logging.warning("Block %s doesn't pass the runs constraints", blockName)
                    self.rejectedWork.append(blockName)
                    continue

                if len(runs) == len(runLumis):
                    # If there is no change in the runs, then we can skip recalculating lumi counts
                    recalculateLumiCounts = False

                if recalculateLumiCounts:
                    # Recalculate effective size of block
                    # We pull out file info, since we don't do this often
                    acceptedLumiCount = 0
                    acceptedEventCount = 0
                    acceptedFileCount = 0
                    fileInfo = dbs.listFilesInBlock(fileBlockName=block['block'])
                    for fileEntry in fileInfo:
                        acceptedFile = False
                        acceptedFileLumiCount = 0
                        for lumiInfo in fileEntry['LumiList']:
                            runNumber = lumiInfo['RunNumber']
                            if runNumber in runs:
                                acceptedFile = True
                                acceptedFileLumiCount += 1
                                acceptedLumiCount += len(lumiInfo['LumiSectionNumber'])
                        if acceptedFile:
                            acceptedFileCount += 1
                            if len(fileEntry['LumiList']) != acceptedFileLumiCount:
                                acceptedEventCount += acceptedFileLumiCount * fileEntry['NumberOfEvents'] / len(fileEntry['LumiList'])
                            else:
                                acceptedEventCount += fileEntry['NumberOfEvents']
                    block[self.lumiType] = acceptedLumiCount
                    block['NumberOfFiles'] = acceptedFileCount
                    block['NumberOfEvents'] = acceptedEventCount
            # save locations
            if task.getTrustSitelists().get('trustlists'):
                self.data[block['block']] = self.sites
            else:
                blockLocations = self.blockLocationRucioPhedex(block['block'])
                self.data[block['block']] = self.cric.PNNstoPSNs(blockLocations)

            # TODO: need to decide what to do when location is no find.
            # There could be case for network problem (no connection to dbs, phedex)
            # or DBS se is not recorded (This will be retried anyway by location mapper)
            if not self.data[block['block']]:
                self.data[block['block']] = ["NoInitialSite"]
            # # No sites for this block, move it to rejected
            #    self.rejectedWork.append(blockName)
            #    continue

            validBlocks.append(block)
        return validBlocks

    def _getBlockSummary(self, dbsObj, datasetPath, blockName):
        """
        Retrieve a summary for this block from both DBS and Rucio. If the block
        has 0 valid files in DBS, or 0 files in Rucio, it is then marked as
        rejected and skipped from the work creation. Otherwise, the DBS summary
        is returned.
        :param dbsObj: instance to the DBS3Reader object
        :param datasetPath: string with the input dataset name
        :param blockName: string with the block name
        :return: either an empty dictionary, or the DBS summary dictionary
        """
        # blocks with 0 valid files should be ignored
        # - ideally they would be deleted but dbs can't delete blocks
        block = dbsObj.getDBSSummaryInfo(dataset=datasetPath, block=blockName)
        if int(block.get('NumberOfFiles', 0)) == 0:
            logging.warning("Block %s being rejected for lack of valid files in DBS to process", blockName)
            self.badWork.append(blockName)
            return dict()
        # blocks with 0 files in Rucio should be ignored as well
        blockRucio = self.rucio.getDID(didName=blockName, dynamic=False)
        if not blockRucio.get('length'):
            logging.warning("Block %s being rejected for lack of files in Rucio to process", blockName)
            self.badWork.append(blockName)
            return dict()
        return block

    def modifyPolicyForWorkAddition(self, inboxElement):
        """
        A block blacklist modifier will be created,
        this policy object will split excluding the blocks in both the spec
        blacklist and the blacklist modified
        """
        # Get the already processed input blocks from the inbox element
        self.blockBlackListModifier = inboxElement.get('ProcessedInputs', [])
        self.blockBlackListModifier.extend(inboxElement.get('RejectedInputs', []))

    def newDataAvailable(self, task, inbound):
        """
            In the case of the block policy, the new data available
            returns True if it finds at least one open block.
        """
        self.initialTask = task
        dbs = self.dbs()
        allBlocks = dbs.listFileBlocks(task.getInputDatasetPath())
        newBlocks = set(allBlocks) - set(self.rejectedWork) - set(self.badWork)
        return bool(newBlocks)

    @staticmethod
    def supportsWorkAddition():
        """
        Block start policy supports continuous addition of work
        """
        return True



#!/usr/bin/env python
"""
_TaskEmulator_

Top level emulator controller

"""

from builtins import object
from WMCore.WMSpec.WMStep import WMStepHelper
import WMCore.WMSpec.Steps.StepFactory as StepFactory


class TaskEmulator(object):
    """
    _TaskEmulator_

    Top Level Emulator that contains a map of all the steps to step emulators
    and be able to run them

    Instantiate with the task to be emulated

    """
    def __init__(self, task):
        self.task = task
        self.emulators = {}

    def getEmulator(self, stepName):
        """
        _getEmulator_

        Retrieve the Emulator for the step name provided, returns None if
        there is no emulator

        """
        return self.emulators.get(stepName, None)


    def addEmulator(self, nodeName, emulatorName):
        """
        _addEmulator_

        Add an Emulator for the node provided, emulatorName is the name
        of the emulator class to be loaded by the Emulator factory

        TODO: Exception handling

        """
        emuInstance = StepFactory.getStepEmulator(emulatorName)
        self.emulators[nodeName] = emuInstance
        return

    def emulateAll(self):
        """
        _emulateAll_

        Traverse all Steps and load up the default Emulator based on
        type.

        """
        for step in self.task.steps().nodeIterator():
            helper = WMStepHelper(step)
            stepType = helper.stepType()
            stepName = helper.name()
            self.addEmulator(stepName, stepType)



    def __call__(self, step):
        """
        _operator(step)_

        Invoke the emulator for the given step

        """



#!/usr/bin/env python
"""
_Delete_

Oracle implementation of Delete
"""




from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.Delete import Delete as MySQLDelete

class Delete(MySQLDelete):
    """
    _Delete_

    Oracle implementation of Delete
    """
    pass



#/usr/bin/env python2.4
"""
_Destroy_

"""




from WMCore.Agent.Database.DestroyAgentBase import DestroyAgentBase

class Destroy(DestroyAgentBase):
    def __init__(self, logger = None, dbi = None):
        """
        _init_

        Call the base class's constructor and add all necessary tables for
        deletion,
        """
        DestroyAgentBase.__init__(self, logger, dbi)



#!/usr/bin/env python
"""
_SetStatus_

MySQL implementation of DBSBufferFiles.SetStatus
"""




from WMCore.Database.DBFormatter import DBFormatter

class SetStatus(DBFormatter):
    sql = "UPDATE dbsbuffer_file SET status = :status WHERE lfn = :lfn"

    def execute(self, lfns, status, conn = None, transaction = None):
        if not isinstance(lfns, list):
            lfns = [lfns]

        bindVars = []
        for lfn in lfns:
            bindVars.append({"lfn": lfn, "status": status})

        result = self.dbi.processData(self.sql, bindVars, conn = conn,
                                      transaction = transaction)
        return



#!/usr/bin/env python
"""
_GetSubTypes_

MySQL implementation of Jobs.GetSubTypes
"""

__all__ = []



import logging

from WMCore.Database.DBFormatter import DBFormatter



class GetSubTypes(DBFormatter):
    """
    Get the current sub types available to WMBS

    """

    sql = """SELECT name FROM wmbs_sub_types"""

    def formatThis(self, result):
        """
        It formats stuff

        """

        res = self.format(result)

        final = []

        for item in res:
            if isinstance(item, list):
                final.extend(item)


        return final

    def execute(self, conn = None, transaction = False):
        """
        Find them types

        """

        result = self.dbi.processData(self.sql, binds = {}, conn = conn, transaction = transaction)


        return self.formatThis(result)



"""
_ReleasePeriodicJob_

MySQL implementation of JobSplitting.ReleasePeriodicJob
"""

import time

from WMCore.Database.DBFormatter import DBFormatter

class ReleasePeriodicJob(DBFormatter):
    """
    _ReleasePeriodicJob_

    Returns either 0 (do not release the next periodic harvesting job)
                or 1 (release the next periodic harvesting job).

    Existence of available files is assumed for this query, because
    a job splitting algorithm is only called if there are
    available files for the subscription.

    To return 1, we need to either have no jobs (jobgroups) or all
    existing jobs need to have finished and their latest state
    change must have occured longer than 'period' seconds ago.
    """

    sql = """SELECT CASE
                      WHEN COUNT(wmbs_jobgroup.id) = 0 THEN 1
                      WHEN COUNT(wmbs_sub_files_acquired.subscription) > 0 THEN 0
                      WHEN COUNT(wmbs_job.id) = 0 THEN 1
                      ELSE 0
                    END
             FROM wmbs_subscription
             LEFT OUTER JOIN wmbs_jobgroup ON
               wmbs_jobgroup.subscription = wmbs_subscription.id
             LEFT OUTER JOIN wmbs_sub_files_acquired ON
               wmbs_sub_files_acquired.subscription = wmbs_subscription.id
             LEFT OUTER JOIN wmbs_job ON
               wmbs_job.jobgroup = wmbs_jobgroup.id AND
               wmbs_job.state_time > :TIMEOUT
             WHERE wmbs_subscription.id = :subscription
             """

    def execute(self, subscription, period, conn = None, transaction = False):

        binds = { 'SUBSCRIPTION' : subscription,
                  'TIMEOUT' : int(time.time()) - period }

        release = self.dbi.processData(self.sql, binds, conn = conn,
                                       transaction = transaction)[0].fetchall()[0][0]

        return (release == 1)



#!/usr/bin/env python
"""
_FailedJobsByWorkflow_

Monitoring DAO classes for Jobs in WMBS
"""
__all__ = []



from WMCore.WMBS.MySQL.Monitoring.FailedJobsByWorkflow \
  import FailedJobsByWorkflow as FailedJobsByWorkflowMySQL

class FailedJobsByWorkflow(FailedJobsByWorkflowMySQL):
    pass



#!/usr/bin/env python
"""
_SiteLocalConfig_

Utility for reading a site local config XML file and converting it
into an object with an API for getting info from it.

"""


from builtins import next, str, object

import os
import logging

from WMCore.Algorithms.ParseXMLFile import xmlFileToNode
from WMCore.Storage.TrivialFileCatalog import tfcFilename, tfcProtocol, readTFC


def loadSiteLocalConfig():
    """
    _loadSiteLocalConfig_

    Runtime Accessor for the site local config.

    Requires that CMS_PATH is defined as an environment variable

    """
    overVarName = "WMAGENT_SITE_CONFIG_OVERRIDE"
    if os.getenv(overVarName, None):
        overridePath = os.getenv(overVarName)
        if os.path.exists(os.getenv(overVarName, None)):
            m = ("Using site-local-config.xml override due to set %s env. variable, "
                 "loading: '%s'" % (overVarName, overridePath))
            logging.log(logging.DEBUG, m)
            config = SiteLocalConfig(overridePath)
            return config
        else:
            msg = "%s env. var. provided but not pointing to an existing file, ignoring." % overVarName
            logging.log(logging.ERROR, msg)

    defaultPath = "$CMS_PATH/SITECONF/local/JobConfig/site-local-config.xml"
    actualPath = os.path.expandvars(defaultPath)
    if os.environ.get("CMS_PATH", None) is None:
        msg = "Unable to find site local config file:\n"
        msg += "CMS_PATH variable is not defined."
        raise SiteConfigError(msg)

    if not os.path.exists(actualPath):
        msg = "Unable to find site local config file:\n"
        msg += actualPath
        raise SiteConfigError(msg)

    config = SiteLocalConfig(actualPath)
    return config


class SiteConfigError(Exception):
    """
    Exception class placeholder

    """
    pass


class SiteLocalConfig(object):
    """
    _SiteLocalConfig_

    Readonly API object for getting info out of the SiteLocalConfig file

    """
    def __init__(self, siteConfigXML):
        self.siteConfigFile = siteConfigXML
        self.siteName = None
        self.eventData = {}

        self.frontierProxies = []
        self.frontierServers = []

        self.localStageOut = {}
        self.fallbackStageOut = []

        self.read()
        return

    def trivialFileCatalog(self):
        """
        _trivialFileCatalog_

        Return an instance of FwkJobRep.TrivialFileCatalog

        """
        tfcUrl = self.localStageOut.get('catalog', None)
        if tfcUrl is None:
            return None
        try:
            tfcFile = tfcFilename(tfcUrl)
            tfcProto = tfcProtocol(tfcUrl)
            tfcInstance = readTFC(tfcFile)
            tfcInstance.preferredProtocol = tfcProto
        except Exception as ex:
            msg = "Unable to load TrivialFileCatalog:\n"
            msg += "URL = %s\n" % tfcUrl
            raise SiteConfigError(msg)
        return tfcInstance


    def localStageOutCommand(self):
        """
        _localStageOutCommand_

        Return the stage out command setting from local-stage-out

        """
        return self.localStageOut['command']

    def localStageOutOption(self):
        """
        _localStageOutOption_

        Return the stage out option setting from local-stage-out
        """
        return self.localStageOut['option']

    def localStageOutPNN(self):
        """
        _localStageOutPNN_

        return the local PhEDExNodeName used for stage out

        """
        return self.localStageOut['phedex-node']


    def read(self):
        """
        _read_

        Load data from SiteLocal Config file and populate this object

        """
        try:
            node = xmlFileToNode(self.siteConfigFile)
        except Exception as ex:
            msg = "Unable to read SiteConfigFile: %s\n" % self.siteConfigFile
            msg += str(ex)
            raise SiteConfigError(msg)

        nodeResult = nodeReader(node)

        if 'siteName' not in nodeResult:
            msg = "Unable to find site name in SiteConfigFile:\n"
            msg += self.siteConfigFile
            raise SiteConfigError(msg)
        if 'catalog' not in nodeResult:
            msg = "Unable to find catalog entry for event data in SiteConfigFile:\n"
            msg += self.siteConfigFile
            raise SiteConfigError(msg)
        if 'localStageOut' not in nodeResult:
            msg = "Error:Unable to find any local-stage-out"
            msg += "information in:\n"
            msg += self.siteConfigFile
            raise SiteConfigError(msg)

        self.siteName             = nodeResult.get('siteName', None)
        self.eventData['catalog'] = nodeResult.get('catalog', None)
        self.localStageOut        = nodeResult.get('localStageOut', [])
        self.fallbackStageOut     = nodeResult.get('fallbackStageOut', [])
        self.frontierServers      = nodeResult.get('frontierServers', [])
        self.frontierProxies      = nodeResult.get('frontierProxies', [])
        return

def coroutine(func):
    """
    _coroutine_

    Decorator method used to prime coroutines

    """
    def start(*args, **kwargs):
        cr = func(*args, **kwargs)
        next(cr)
        return cr
    return start



def nodeReader(node):
    """
    _nodeReader_

    Given a node, see if we can find what we're looking for
    """
    processSiteInfo = {
        'event-data': processEventData(),
        'local-stage-out': processLocalStageOut(),
        'calib-data': processCalibData(),
        'fallback-stage-out': processFallbackStageOut()
        }

    report = {}
    sProcess = processSite(processSiteInfo)
    processor = processNode(sProcess)
    processor.send((report, node))

    return report

@coroutine
def processNode(target):
    """
    Starts at the top of the tree and finds the site
    """
    while True:
        report, node = (yield)
        for subnode in node.children:
            if subnode.name == 'site-local-config':
                for child in subnode.children:
                    if child.name == 'site':
                        target.send((report, child))


@coroutine
def processSite(targets):
    """
    Process the site tree in a config.

    """

    while True:
        report, node = (yield)
        #Get the name first
        report['siteName'] = node.attrs.get('name', None)
        for subnode in node.children:
            if subnode.name == 'event-data':
                targets['event-data'].send((report, subnode))
            elif subnode.name == 'calib-data':
                targets['calib-data'].send((report, subnode))
            elif subnode.name == 'local-stage-out':
                targets['local-stage-out'].send((report, subnode))
            elif subnode.name == 'fallback-stage-out':
                targets['fallback-stage-out'].send((report, subnode))



@coroutine
def processEventData():
    """
    Process eventData in a site

    """

    while True:
        report, node = (yield)
        for subnode in node.children:
            if subnode.name == 'catalog':
                report['catalog'] = str(subnode.attrs.get('url', None))

@coroutine
def processLocalStageOut():
    """
    Find the local-stage-out directory

    """

    while True:
        report, node = (yield)
        localReport = {}
        for subnode in node.children:
            if subnode.name in ['phedex-node', 'command', 'option']:
                localReport[subnode.name] = subnode.attrs.get('value', None)
            elif subnode.name == 'catalog':
                localReport[subnode.name] = subnode.attrs.get('url', None)
        report['localStageOut'] = localReport

@coroutine
def processFallbackStageOut():
    """
    Find the fallback stage out directory

    """

    while True:
        report, node = (yield)
        localReport = {}
        for subnode in node.children:
            if subnode.name in ['phedex-node', 'command', 'option', 'lfn-prefix']:
                localReport[subnode.name] = subnode.attrs.get('value', None)
        report['fallbackStageOut'] = [localReport]


@coroutine
def processCalibData():
    """
    Process calib-data

    """

    while True:
        report, node = (yield)

        frontierProxies = []
        frontierServers = []
        for subnode in node.children:
            if subnode.name == "frontier-connect":
                for frontierSubnode in subnode.children:
                    subNodeUrl = frontierSubnode.attrs.get("url", None)

                    if frontierSubnode.name == "proxy":
                        frontierProxies.append(subNodeUrl)
                    elif frontierSubnode.name == "server":
                        frontierServers.append(subNodeUrl)

        report["frontierProxies"] = frontierProxies
        report["frontierServers"] = frontierServers



from __future__ import (division, print_function)

from time import time
from WMCore.REST.CherryPyPeriodicTask import CherryPyPeriodicTask
from WMCore.WorkQueue.WorkQueue import globalQueue

class CleanUpTask(CherryPyPeriodicTask):

    def __init__(self, rest, config):

        super(CleanUpTask, self).__init__(config)
        self.globalQ = globalQueue(logger=self.logger, **config.queueParams)

    def setConcurrentTasks(self, config):
        """
        sets the list of function reference for concurrent tasks
        """
        self.concurrentTasks = [{'func': self.cleanUpAndSyncCanceledElements, 'duration': config.cleanUpDuration}]

    def cleanUpAndSyncCanceledElements(self, config):
        """
        1. deleted the wqe in end states
        2. synchronize cancelled elements.
        We can also make this in the separate thread
        """
        tStart = time()
        self.globalQ.performQueueCleanupActions(skipWMBS=True)
        self.logger.info("%s executed in %.3f secs.", self.__class__.__name__, time() - tStart)
        return



#!/usr/bin/env python
# encoding: utf-8
"""
CouchService.py

Created by Dave Evans on 2010-04-20.
Copyright (c) 2010 Fermilab. All rights reserved.
"""
from builtins import object
from time import time

import WMCore.Database.CouchUtils as CouchUtils


class CouchService(object):
    def __init__(self, **options):
        super(CouchService, self).__init__()
        self.options = {}
        self.options.update(options)
        self.url = options.get('url', None)
        self.database = options.get('database', None)
        self.server = None
        self.couchdb = None

    def listFilesets(self, collectionInstance):
        """
        _listFilesets_

        List filesets for the collection instance provided.
        """
        collectionInstance.populate()

        for fileset in collectionInstance["filesets"]:
            yield fileset

    @CouchUtils.connectToCouch
    def removeFilesetsByCollectionName(self, collectionName):
        """
        _removeFilesetsByCollectionName_

        Remove all the collections matching certain collection
        name.
        """
        result = self.couchdb.loadView("ACDC", "byCollectionName", options={"reduce": False}, keys=[collectionName])
        for entry in result["rows"]:
            self.couchdb.queueDelete(entry["value"])
        return self.couchdb.commit()

    @CouchUtils.connectToCouch
    def removeOldFilesets(self, expirationDays):
        """
        _removeOldFilesets_

        Remove filesets older than certain date defined
        in expirationDays (in days).
        """
        cutoutPoint = time() - (expirationDays * 3600 * 24)
        result = self.couchdb.loadView("ACDC", "byTimestamp", {"endkey": cutoutPoint})
        count = 0
        for entry in result["rows"]:
            self.couchdb.queueDelete(entry["value"])
            count += 1
        self.couchdb.commit()
        return count

    @CouchUtils.connectToCouch
    def listCollectionNames(self):
        options = {'reduce': True, 'group_level': 1, 'stale': "update_after"}
        result = self.couchdb.loadView("ACDC", "byCollectionName", options)
        collectionNames = []
        for row in result["rows"]:
            collectionNames.append(row["key"])
        return collectionNames



#!/usr/bin/env python
"""
_JobsByState_

Monitoring DAO classes for Jobs in WMBS
"""
__all__ = []



from WMCore.WMBS.MySQL.Monitoring.DefaultFormatter import DefaultFormatter

# Job Related Query
class JobsByState(DefaultFormatter):

    #TO check what else is needed for return item
    sql = """SELECT wmbs_job.id AS jobID, wmbs_job.name AS job_name, wmbs_job_state.name AS job_state
             FROM wmbs_job
             INNER JOIN wmbs_job_state ON wmbs_job.state=wmbs_job_state.id ORDER BY state"""



"""
_ListSubscriptions_

Oracle implementation of DBS3Buffer.ListSubscriptions

Created on May 2, 2013

@author: dballest
"""

from WMComponent.DBS3Buffer.MySQL.ListSubscriptions import ListSubscriptions as MySQLListSubscriptions

class ListSubscriptions(MySQLListSubscriptions):
    pass



#!/usr/bin/env python
"""
_UpdateJobs_

MySQL implementation for updating jobs
"""

from WMCore.Database.DBFormatter import DBFormatter


class UpdateJobs(DBFormatter):
    """
    _UpdateJobs_

    Update jobs with new values
    """

    sql = """UPDATE bl_runjob SET wmbs_id = :jobid, grid_id = :gridid,
               bulk_id = :bulkid, status_time = :status_time,
               sched_status = (SELECT id FROM bl_status WHERE name = :status),
               retry_count = :retry_count,
               user_id = (SELECT id FROM wmbs_users WHERE cert_dn = :owner AND group_name = :usergroup AND role_name = :userrole)
               WHERE id = :id
               """

    def execute(self, jobs, conn=None, transaction=False):
        """
        _execute_

        Update jobs with new values.
        Mostly a maintenance script
        """

        if len(jobs) == 0:
            return

        binds = []
        for job in jobs:
            binds.append({'jobid': job['jobid'], 'gridid': job.get('gridid', None), 'bulkid': job.get('bulkid', None),
                          'status': job.get('status', None), 'retry_count': job['retry_count'], 'id': job['id'],
                          'status_time': job.get('status_time', None), 'owner': job['userdn'],
                          'usergroup': job['usergroup'], 'userrole': job['userrole']})

        self.dbi.processData(self.sql, binds, conn=conn, transaction=transaction)

        return



#!/usr/bin/env python
"""
_SetLastUpdate_

MySQL implementation of Fileset.SetLastUpdate
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter
import time

class SetLastUpdate(DBFormatter):
    sql = "UPDATE wmbs_fileset SET last_update = :p_1 WHERE name = :p_2"

    def execute(self, fileset = None, timeUpdate = int(time.time()) , \
                conn = None, transaction = False):
        bindVars = {"p_1": timeUpdate, "p_2": fileset}
        self.dbi.processData(self.sql, bindVars, conn = conn,
                            transaction = transaction)
        return



#!/usr/bin/env python
"""
_ListWorkloadsForTaskSite_

List the task name and number of jobs running for a given site and subscription
type.
"""




from WMCore.Database.DBFormatter import DBFormatter

class ListWorkloadsForTaskSite(DBFormatter):
    sql = """SELECT wmbs_workflow.task AS task, COUNT(wmbs_job.id) AS running FROM wmbs_workflow
               INNER JOIN wmbs_subscription ON
                 wmbs_workflow.id = wmbs_subscription.workflow
               INNER JOIN wmbs_sub_types ON
                 wmbs_subscription.subtype = wmbs_sub_types.id
               INNER JOIN wmbs_jobgroup ON
                 wmbs_subscription.id = wmbs_jobgroup.subscription
               INNER JOIN wmbs_job ON
                 wmbs_jobgroup.id = wmbs_job.jobgroup
               INNER JOIN wmbs_job_state ON
                 wmbs_job.state = wmbs_job_state.id
               INNER JOIN wmbs_location ON
                 wmbs_job.location = wmbs_location.id
             WHERE wmbs_job_state.name = 'executing' AND
                   wmbs_sub_types.name = :task_type AND
                   wmbs_location.site_name = :site_name
             GROUP BY wmbs_workflow.task"""

    def execute(self, taskType, siteName, conn = None, transaction = False):
        results = self.dbi.processData(self.sql, {"task_type": taskType,
                                                  "site_name": siteName},
                                       conn = conn, transaction = transaction)
        return self.formatDict(results)



from __future__ import (division, print_function)

import json

from WMCore.Services.Service import Service


class WMArchive(Service):
    """
    This is skelton class which need be implemented.
    """
    def __init__(self, url, header = {}):
        """
        responseType will be either xml or json
        """

        httpDict = {}
        # url is end point
        httpDict['endpoint'] = "%s/data" % url

        # cherrypy converts request.body to params when content type is set
        # application/x-www-form-urlencodeds
        httpDict.setdefault("content_type", 'application/json')
        httpDict.setdefault('cacheduration', 0)
        httpDict.setdefault("accept_type", "application/json")
        httpDict.update(header)
        self.encoder = json.dumps
        Service.__init__(self, httpDict)



    def archiveData(self, data):
        return self["requests"].post('', {'data': data})[0]['result']



#!/usr/bin/env python
"""
_GetAvailableFilesNoLocations_

Retrieve the IDs of available files without their locations.
"""

from WMCore.WMBS.MySQL.Subscriptions.GetAvailableFiles import GetAvailableFiles as GetAvailableFilesMySQL

class GetAvailableFilesNoLocations(GetAvailableFilesMySQL):
    sql = """SELECT wmbs_sub_files_available.fileid AS fileid FROM wmbs_sub_files_available
             WHERE wmbs_sub_files_available.subscription = :subscription"""



#!/usr/bin/env python
"""
_StageOut_

Template for a StageOut Step

"""

from WMCore.WMSpec.Steps.Template import Template
from WMCore.WMSpec.Steps.Template import CoreHelper
from WMCore.WMSpec.ConfigSectionTree import nodeName

class StageOutStepHelper(CoreHelper):
    """
    _StageOutStepHelper_

    Add API calls and helper methods to the basic WMStepHelper to specialise
    for StageOut tasks

    """
    def addFile(self, infile, outfile, stageOut = True):
        """
            Enqueues(sp?) a file to the StageOut step
            infile must be a LOCAL FILENAME. No file:/// in front
            outfile must be a LFN which will be mapped to the remote
        """
        target = self.data.files.section_("file%i" % self.data.filecount)
        target.input         = infile
        target.output        = outfile
        target.stageOut      = stageOut
        self.data.filecount += 1

    def disableRetries(self):
        """
            handy for testing, without the 10 minute retry loop
        """
        self.data.retryCount = 1
        self.data.retryDelay = 0

    def disableStraightToMerge(self):
        """
        _disableStraightToMerge_

        Disable straight to merge for this step.
        """
        if hasattr(self.data.output, "minMergeSize"):
            delattr(self.data.output, "minMergeSize")

        return

    def disableStraightToMergeForOutputModules(self, outputModules):
        """
        _disableStraightToMergeForOutputModules_

        Disable straight to merge only for these output modules.
        """
        self.data.output.forceUnmergedOutputs = outputModules

        return

    def setMinMergeSize(self, minMergeSize, maxMergeEvents):
        """
        _setMinMergeSize_

        Set the mininum size for promoting a file to merged status.
        """
        self.data.output.minMergeSize = minMergeSize
        self.data.output.maxMergeEvents = maxMergeEvents
        return

    def minMergeSize(self):
        """
        _minMergeSize_

        Retrieve the minimum size for promoting a file to merged status.  If
        straight to merge is disabled -1 will be returned.
        """
        return getattr(self.data.output, "minMergeSize", -1)

class StageOut(Template):
    """
    _StageOut_

    Tools for creating a template StageOut Step

    """

    def install(self, step):
        step.stepType = "StageOut"
        step.section_("files")
        step.filecount = 0
        step.retryCount = 3
        step.retryDelay = 300


    def helper(self, step):
        """
        _helper_

        Wrap the WMStep provided in the CMSSW helper class that
        includes the ability to add and manipulate the details
        of a CMSSW workflow step

        """
        return StageOutStepHelper(step)



"""
_GetNumberOfJobsForWorkflowTaskStatus_

Oracle implementation of GetNumberOfJobsForWorkflowTaskStatus

Created on Apr 17, 2013

@author: dballest
"""
__all__ = []

from WMCore.WMBS.MySQL.Jobs.GetNumberOfJobsForWorkflowTaskStatus import GetNumberOfJobsForWorkflowTaskStatus \
    as MySQLGetNumberOfJobsForWorkflowTaskStatus

class GetNumberOfJobsForWorkflowTaskStatus(MySQLGetNumberOfJobsForWorkflowTaskStatus):
    pass



#!/usr/bin/env python
"""
_New_

MySQL implementation of JobGroup.New
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class New(DBFormatter):
    sql = """INSERT INTO wmbs_jobgroup (subscription, guid, output,
             last_update) VALUES (:subscription, :guid, :output,
             unix_timestamp())"""

    def execute(self, uid, subscription = None, output = None, conn = None,
                transaction = False):
        binds = self.getBinds(subscription = subscription, guid = uid,
                              output = output)
        result = self.dbi.processData(self.sql, binds, conn = conn,
                                      transaction = transaction)
        return



#!/usr/bin/env python
"""
_GetBulkRunLumi_

Oracle implementation of Files.GetBulkRunLumi
"""

from WMCore.WMBS.MySQL.Files.GetBulkRunLumi import GetBulkRunLumi as MySQLGetBulkRunLumi

class GetBulkRunLumi(MySQLGetBulkRunLumi):
    sql = """SELECT flr.run AS run, flr.lumi AS lumi, flr.fileid AS id
               FROM wmbs_file_runlumi_map flr
               WHERE flr.fileid = :id"""



#!/usr/bin/env python
"""
_TaskMaker_

This should take an incoming WMWorkload, break it into tasks
and then for each task do the jobs necessary for the task
to start as a proper job.

"""
from __future__ import print_function

from builtins import object
import logging
import os.path
import threading

from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMCore.WMRuntime.SandboxCreator import SandboxCreator
from WMCore.WMSpec.WMWorkload import WMWorkload, WMWorkloadHelper


# The class should carry the following variables:
#  self.workload = WMSpec.WMWorkloadHelper object, loaded by loadWorkload
#  self.workdir  = string carrying the name of an open working directory, loaded by loadWorkdir
#  self.workflowDict = dictionary of WMBS.Workflows keyed by task name;
#  self.subDict  = dictionary of WMBS.Subscriptions keyed by task name;
#


class TaskMaker(object):
    """
    Class for separating and starting all tasks in a WMWorkload

    """

    def __init__(self, workload=None, workdir=None):
        self.workload = None
        self.workdir = None
        self.skipSubscription = False
        self.workflowDict = {}
        self.subDict = {}
        self.owner = "unknown"
        self.owner_dn = "unknown"

        myThread = threading.currentThread()
        myThread.logger = logging.getLogger()

        self.loadWorkload(workload)
        self.loadWorkdir(workdir)

        return

    def loadWorkload(self, inputWorkload):
        """
        If workload is sane, then use it

        """

        if inputWorkload is None:
            self.workload = None
        if isinstance(inputWorkload, WMWorkload):
            self.workload = WMWorkloadHelper(inputWorkload)
            return
        if isinstance(inputWorkload, WMWorkloadHelper):
            self.workload = inputWorkload
            return

        if not os.path.exists(inputWorkload):
            raise Exception('Could not find %s in local file system' % (str(inputWorkload)))

        testWorkload = WMWorkloadHelper(WMWorkload("workload"))
        testWorkload.load(inputWorkload)

        self.workload = testWorkload

        return

    def loadWorkdir(self, inputWorkdir):
        """
        If workdir is sane, use it

        """

        if inputWorkdir is None:
            self.workdir = None
            return

        if not os.path.isdir(inputWorkdir):
            try:
                os.mkdir(inputWorkdir)
            except Exception as ex:
                print('Caught exception %s in creating workdir %s' % (str(ex), inputWorkdir))
                raise Exception(ex)

        self.workdir = inputWorkdir

        return

    def processWorkload(self):
        """
        Split the workload into tasks, and finish all the process for the task

        """

        if self.workdir is None or self.workload is None:
            raise Exception(
                'Failure in processing workload %s in directory %s' % (str(self.workload), str(self.workdir)))

        if hasattr(self.workload, 'owner'):
            self.owner = self.workload.owner
        if hasattr(self.workload, 'owner_dn'):
            self.owner_dn = self.workload.owner_dn

        for toptask in self.workload.taskIterator():
            # for each task, build sandbox, register, and subscribe
            for task in toptask.taskIterator():
                if task.name() in self.workflowDict:
                    raise Exception(
                        'Duplicate task name for workload %s, task %s' % (self.workload.name(), task.name()))

                if not self.skipSubscription:
                    self.subscribeWMBS(task)

        sandboxCreator = SandboxCreator()
        sandboxCreator.makeSandbox(self.workdir, self.workload)
        logging.info('Done processing workload %s' % (self.workload.name()))

        return True

    def createWorkflow(self, task):
        """
        Register job into WMBS for each task through Workflows

        """

        specURL = self.getWorkflowURL(task)

        fileSet = Fileset(name=self.getFilesetName(task), is_open=True)
        fileSet.create()

        taskFlow = Workflow(spec=specURL, owner=self.owner, dn=self.owner_dn,
                            name=self.getWorkflowName(task), task=task.name())
        taskFlow.create()

        self.workflowDict[task.name()] = taskFlow

        # Insert workflow into task
        setattr(task.data.input.WMBS, 'WorkflowSpecURL', specURL)

        # If the job is a merge job
        # Find the task it merges from
        # Then find the workflow for that task and assign it an output
        if hasattr(task.inputReference(), 'outputModule'):
            dummyStepName = task.inputReference().inputStep.split('/')[-1]
            taskName = task.inputReference().inputStep.split('/')[-2]
            outputModule = task.inputReference().outputModule
            if taskName not in self.workflowDict:
                raise Exception('I am being asked to chain output for a task %s which does not yet exist' % taskName)
            outputWorkflow = self.workflowDict[taskName]
            outputWorkflow.addOutput(outputModule, fileSet)

        logging.info('Registered workflow for step %s' % (task.name()))

        return taskFlow, fileSet

    def subscribeWMBS(self, task):
        """
        Create a subscription for each task

        """

        workFlow, fileSet = self.createWorkflow(task)

        workFlow.load()
        fileSet.load()

        subType = ''
        if task.name() == 'Processing' or task.name() == 'Production':
            subType = 'Processing'
        elif task.name() == 'Merge':
            subType = 'Merge'

        newSub = Subscription(fileset=fileSet, workflow=workFlow, split_algo='FileBased', type=subType)
        newSub.create()

        # Add subscription to dictionary
        self.subDict[task.name()] = newSub

        # Add subscription id to task
        setattr(task.data.input.WMBS, 'Subscription', newSub['id'])

        if not newSub.exists() >= 0:
            raise Exception("ERROR: Subscription does not exist after it was created")

        logging.info('Created subscription for task %s' % (task.name()))

        return

    def getFilesetName(self, task):
        """
        Create a unique name for a fileset

        """

        filesetName = '%s/Fileset-%s-%s' % (self.workdir, self.workload.name(), task.name())

        return filesetName

    def getWorkflowURL(self, task):
        """
        Create a unique name and location for each workflow

        """

        workflowURL = '%s/%s' % (self.workdir, self.getWorkflowName(task))

        return workflowURL

    def getWorkflowName(self, task):
        """
        Create a unique name for each workflow

        """

        workflowName = 'Workflow-%s-%s' % (self.workload.name(), task.name())

        return workflowName



#!/usr/bin/env python
"""
_ScriptFactory_

WMFactory for instantiating runtime implementations of the ScriptInterface

"""

from builtins import str
from WMCore.WMFactory import WMFactory
from WMCore.WMException import WMException

class ScriptFactoryException(WMException):
    """
    _ScriptFactortyException_

    Exception for missing objects or problems

    """
    pass



class ScriptFactory(WMFactory):
    """
    _ScriptFactory_

    Instantiate a WMFactory instance with the appropriate namespace

    """
    def __init__(self):
        WMFactory.__init__(self, self.__class__.__name__,
                           "WMCore.WMRuntime.Scripts")

_ScriptFactory = ScriptFactory()

def getScript(scriptName):
    """
    _getScript_

    factory method to return a ScriptInterface impl instance based on the
    name provided

    """
    try:
        return _ScriptFactory.loadObject(scriptName)
    except WMException as wmEx:
        msg = "ScriptFactory Unable to load Object: %s" % scriptName
        raise ScriptFactoryException(msg)
    except Exception as ex:
        msg = "Error creating object %s in ScriptFactory:\n" % scriptName
        msg += str(ex)
        raise ScriptFactoryException(msg)



#!/usr/bin/env python
"""
_LoadFromName_

MySQL implementation of Workflow.LoadFromName
"""

from WMCore.Database.DBFormatter import DBFormatter


class LoadFromName(DBFormatter):
    sql = """SELECT * FROM wmbs_workflow WHERE wmbs_workflow.name = :workflow
             ORDER BY task"""

    def execute(self, workflow, conn=None, transaction=False):
        result = self.dbi.processData(self.sql, {"workflow": workflow},
                                      conn=conn, transaction=transaction)
        return self.formatDict(result)



#!/usr/bin/env python
"""
_LoadJobs_

MySQL implementation of JobGroup.LoadJobs
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class LoadJobs(DBFormatter):
    sql = "SELECT id FROM wmbs_job WHERE jobgroup = :jobgroup"

    def formatDict(self, results):
        """
        _formatDict_

        Cast the id to be an integer as formatDict() turns all results into
        strings.
        """
        formattedResults = DBFormatter.formatDict(self, results)

        for formattedResult in formattedResults:
            formattedResult["id"] = int(formattedResult["id"])

        return formattedResults

    def execute(self, jobgroup, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, {"jobgroup": jobgroup},
                                      conn = conn, transaction = transaction)
        return self.formatDict(result)



#!/usr/bin/env python
"""
_UpdateHyperNewsName_

Oracle implementation of Users.UpdateHyperNewsName
"""

from WMCore.WMBS.MySQL.Users.UpdateHyperNewsName import UpdateHyperNewsName as MySQLUpdateHyperNewsName

class UpdateHyperNewsName(MySQLUpdateHyperNewsName):
    pass



#!/usr/bin/env python
# pylint: disable=W0212
# W0212: Names are not accessible for WMTask objects, etc.
"""
_CMSSW_

Diagnostic implementation for a CMSSW job


"""
from __future__ import print_function

from builtins import range, str

import logging
import os.path
import socket

from Utils import FileTools
from WMCore.FwkJobReport.Report import FwkJobReportException
from WMCore.WMSpec.Steps.Diagnostic import Diagnostic, DiagnosticHandler

# strip the lines from the log for the report
DEFAULT_TAIL_LINES_FROM_LOG = 25


class Exit127(DiagnosticHandler):
    """
    Handle non-existant executable

    """

    def __call__(self, errCode, executor, **args):
        msg = "Executable Not Found at host: %s" % socket.getfqdn()
        if args.get('ExceptionInstance', False):
            msg += str(args.get('ExceptionInstance'))
        executor.report.addError(executor.stepName,
                                 50110, "ExecutableNotFound", msg)


class Exit126(DiagnosticHandler):
    """
    Handle bad permissions

    """

    def __call__(self, errCode, executor, **args):
        msg = "Executable permissions not executable at host: %s" % socket.getfqdn()
        if args.get('ExceptionInstance', False):
            msg += str(args.get('ExceptionInstance'))
        executor.report.addError(executor.stepName,
                                 50111, "ExecutableBadPermissions", msg)


class Exit50513(DiagnosticHandler):
    """
    Handle SCRAM script failure

    """

    def __call__(self, errCode, executor, **args):
        """
        Added for Steve to handle SCRAM script failure

        Must fail job (since SCRAM didn't run)

        """
        msg = "SCRAM scripts failed to run!\n"
        if args.get('ExceptionInstance', False):
            msg += str(args.get('ExceptionInstance'))

        executor.report.addError(executor.stepName,
                                 50513, "SCRAMScriptFailure", msg)

        # Then mark the job as failed
        if executor.report.report.status == 0:
            executor.report.report.status = 1


class CMSDefaultHandler(DiagnosticHandler):
    """
    _CMSDefaultHandler_

    Default handler for miscellaneous CMSSW errors.
    """

    def __call__(self, errCode, executor, **args):
        logging.critical("%s Diagnostic Handler invoked", self.__class__.__name__)
        msg = "Error in CMSSW: %s\n" % (errCode)
        jobRepXml = os.path.join(executor.step.builder.workingDir,
                                 executor.step.output.jobReport)

        excepInst = args.get('ExceptionInstance', None)

        description = "Misc. CMSSW error"
        if excepInst:
            if hasattr(excepInst, 'detail'):
                description = excepInst.detail
            msg += str(excepInst)

        if os.path.exists(jobRepXml):
            # job report XML exists, load the exception information from it
            try:
                self.parse(executor, jobRepXml)
            except FwkJobReportException:    
                # Job report is bad, the parse already puts a 50115 in the file    
                pass

            reportStep = executor.report.retrieveStep(executor.stepName)
            reportStep.status = errCode

        # Grab stderr log from CMSSW
        errLog = os.path.join(os.path.dirname(jobRepXml),
                              '%s-stderr.log' % (executor.stepName))
        outLog = os.path.join(os.path.dirname(jobRepXml),
                              '%s-stdout.log' % (executor.stepName))

        if os.path.exists(errLog):
            logTail = FileTools.tail(errLog, DEFAULT_TAIL_LINES_FROM_LOG)
            msg += '\n Adding last %s lines of CMSSW stderr:\n' % DEFAULT_TAIL_LINES_FROM_LOG
            msg += logTail
        if os.path.exists(outLog):
            logTail = FileTools.tail(outLog, DEFAULT_TAIL_LINES_FROM_LOG)
            msg += '\n Adding last %s lines of CMSSW stdout:\n' % DEFAULT_TAIL_LINES_FROM_LOG
            msg += logTail

        executor.report.addError(executor.stepName,
                                 errCode, description, msg)

        return


class CMSRunHandler(DiagnosticHandler):
    """
    _CMSRunHandler_

    Base util for a cmsRun bork, check job report exists and contains appropriate error

    """

    def __init__(self, code, desc):
        DiagnosticHandler.__init__(self)
        self.code = code
        self.desc = desc

    def __call__(self, errCode, executor, **args):
        print("%s Diagnostic Handler invoked" % self.__class__.__name__)
        msg = "Exit %s: %s Exception from cmsRun" % (self.code, self.desc)
        jobRepXml = os.path.join(executor.step.builder.workingDir,
                                 executor.step.output.jobReport)

        if os.path.exists(jobRepXml):
            # job report XML exists, load the exception information from it
            try:
                self.parse(executor, jobRepXml)
            except FwkJobReportException:    
                # Job report is bad, the parse already puts a 50115 in the file    
                pass

            reportStep = executor.report.retrieveStep(executor.stepName)
            reportStep.status = self.code

        errLog = os.path.join(os.path.dirname(jobRepXml),
                              '%s-stderr.log' % (executor.stepName))
        outLog = os.path.join(os.path.dirname(jobRepXml),
                              '%s-stdout.log' % (executor.stepName))

        if os.path.exists(errLog):
            logTail = FileTools.tail(errLog, DEFAULT_TAIL_LINES_FROM_LOG)
            msg += '\n Adding last %s lines of CMSSW stderr:\n' % DEFAULT_TAIL_LINES_FROM_LOG
            msg += logTail
        if os.path.exists(outLog):
            logTail = FileTools.tail(outLog, DEFAULT_TAIL_LINES_FROM_LOG)
            msg += '\n Adding last %s lines of CMSSW stdout:\n' % DEFAULT_TAIL_LINES_FROM_LOG
            msg += logTail

        # make sure the report has the error in it
        errSection = getattr(executor.report.report, "errors", None)
        if errSection is None:
            executor.report.addError(executor.stepName,
                                     self.code, self.desc, msg)
        else:
            if not hasattr(errSection, self.desc):
                executor.report.addError(executor.stepName,
                                         self.code, self.desc, msg)

        print(executor.report.report.errors)
        return


class EDMExceptionHandler(DiagnosticHandler):
    """
    _EDMExceptionHandler_

    Handle unspecified non-zero exit code by checking for a job report file containing an
    EDM error code and report that.

    If the job report isnt there, thats a specific failure

    """

    def __call__(self, errCode, executor, **args):
        """
        _operator()_

        Look for the XML job report, try and read it and extract the error information from it

        """
        jobRepXml = os.path.join(executor.step.builder.workingDir,
                                 executor.step.output.jobReport)

        errLog = os.path.join(os.path.dirname(jobRepXml),
                              '%s-stderr.log' % (executor.stepName))
        outLog = os.path.join(os.path.dirname(jobRepXml),
                              '%s-stdout.log' % (executor.stepName))

        addOn = '\n'
        if os.path.exists(errLog):
            logTail = FileTools.tail(errLog, 10)
            addOn += '\nAdding last ten lines of CMSSW stderr:\n'
            addOn += logTail
        else:
            logging.error("No stderr from CMSSW")
            logging.error(os.listdir(os.path.basename(jobRepXml)))

        if os.path.exists(outLog):
            logTail = FileTools.tail(outLog, DEFAULT_TAIL_LINES_FROM_LOG)
            msg = '\n Adding last %s lines of CMSSW stdout:\n' % DEFAULT_TAIL_LINES_FROM_LOG
            msg += logTail

        # Add the error we were sent
        ex = args.get('ExceptionInstance', None)
        executor.report.addError(executor.stepName,
                                 errCode, "CMSSWStepFailure", msg + str(ex))

        if not os.path.exists(jobRepXml):
            # no report => Error
            msg = "No Job Report Found: %s" % jobRepXml
            executor.report.addError(executor.stepName,
                                     50115, "MissingJobReport", msg)
            return

        # job report XML exists, load the exception information from it
        try:
            self.parse(executor, jobRepXml)
        except FwkJobReportException:    
            # Job report is bad, the parse already puts a 50115 in the file    
            pass

        # make sure the report has the error in it
        errSection = getattr(executor.report.report, "errors", None)
        if errSection is None:
            msg = "Job Report contains no error report, but cmsRun exited non-zero: %s" % errCode
            msg += addOn
            executor.report.addError(executor.stepName,
                                     50116, "MissingErrorReport", msg)
            return

        else:
            # check exit code in report is non zero
            if executor.report.report.status == 0:
                msg = "Job Report contains no error report, but cmsRun exited non-zero: %s" % errCode
                msg += addOn
                executor.report.addError(executor.stepName,
                                         50116, "MissingErrorReport", msg)

            else:
                msg = "Adding extra error in order to hold error report"
                msg += addOn
                executor.report.addError(executor.stepName,
                                         99999, "ErrorLoggingAddition", msg)
        return


class CMSSW(Diagnostic):
    """
    Master diagnostic class

    Sets the default handler.
    """

    def __init__(self):
        Diagnostic.__init__(self)
        self.handlers[50513] = Exit50513()
        self.handlers[126] = Exit126()
        self.handlers[127] = Exit127()
        self.handlers[65] = CMSRunHandler(8001, "CMSException")
        self.handlers[66] = CMSRunHandler(8002, "StdException")
        self.handlers[67] = CMSRunHandler(8003, "UnknownException")
        self.handlers[68] = CMSRunHandler(8004, "StdBadAlloc")
        self.handlers[88] = CMSRunHandler(7000, "CommandLineProcessing")
        self.handlers[89] = CMSRunHandler(7001, "ConfigFileNotFound")
        self.handlers[90] = CMSRunHandler(7002, "ConfigFileReadError")

        self.defaultHandler = CMSDefaultHandler()

        # for all the exception codes between 1 and 225, use a default that attempts to read the code
        # from the job report
        catchAll = EDMExceptionHandler()
        for x in range(0, 255):
            if x not in self.handlers:
                self.handlers.__setitem__(x, catchAll)



#!/usr/bin/env python
"""
_UserFileCache_

API for UserFileCache service
"""

from builtins import str, bytes

import os
import json
import shutil
import hashlib
import tarfile
import tempfile

from Utils.Utilities import encodeUnicodeToBytesConditional
from Utils.PythonVersion import PY3

from WMCore.Services.Service import Service


def calculateChecksum(tarfile_, exclude=None):
    """
    Calculate the checksum of the tar file in input.

    The tarfile_ input parameter could be a string or a file object (anything compatible
    with the fileobj parameter of tarfile.open).

    The exclude parameter could be a list of strings, or a callable that takes as input
    the output of  the list of tarfile.getmembers() and return a list of strings.
    The exclude param is interpreted as a list of files that will not be taken into consideration
    when calculating the checksum.

    The output is the checksum of the tar input file.

    The checksum is calculated taking into consideration the names of the objects
    in the tarfile (files, directories etc) and the content of each file.

    Each file is exctracted, read, and then deleted right after the input is passed
    to the hasher object. The file is read in chuncks of 4096 bytes to avoid memory
    issues.
    """
    if exclude==None: #[] is a dangerous value for a param
        exclude = []

    hasher = hashlib.sha256()

    ## "massage" out the input parameters
    if isinstance(tarfile_, (str, bytes)):
        tar = tarfile.open(tarfile_, mode='r')
    else:
        tar = tarfile.open(fileobj=tarfile_, mode='r')

    if exclude and hasattr(exclude, '__call__'):
        excludeList = exclude(tar.getmembers())
    else:
        excludeList = exclude


    tmpDir = tempfile.mkdtemp()
    try:
        for tarmember in tar:
            if tarmember.name in excludeList:
                continue
            hasher.update(encodeUnicodeToBytesConditional(tarmember.name, condition=PY3))
            if tarmember.isfile() and tarmember.name.split('.')[-1]!='pkl':
                tar.extractall(path=tmpDir, members=[tarmember])
                fn = os.path.join(tmpDir, tarmember.name)
                with open(fn, 'rb') as fd:
                    while True:
                        buf = fd.read(4096)
                        if not buf:
                            break
                        hasher.update(buf)
                os.remove(fn)
    finally:
        #never leave tmddir around
        shutil.rmtree(tmpDir)
    checksum = hasher.hexdigest()

    return checksum


class UserFileCache(Service):
    """
    API for UserFileCache service
    """
    # Should be filled out with other methods: download, exists

    def __init__(self, mydict=None):
        if mydict==None: #dangerous {} default value
            mydict = {}
        mydict['endpoint'] =  mydict.get('endpoint', 'https://cmsweb.cern.ch/crabcache/')
        Service.__init__(self, mydict)
        self['requests']['accept_type'] = 'application/json'

        if 'proxyfilename' in mydict:
            #in case there is some code I have not updated in ticket #3780. Should not be required... but...
            self['logger'].warning('The UserFileCache proxyfilename parameter has been replace with the more'
                                   ' general (ckey/cert) pair.')

    def downloadLog(self, fileName, output):
        """
        """
        url = self['endpoint'] + 'logfile?name=%s' % os.path.split(fileName)[1]

        self['logger'].info('Fetching URL %s' % url)
        fileName, dummyHeader = self['requests'].downloadFile(output, str(url)) #unicode broke pycurl.setopt
        self['logger'].debug('Wrote %s' % output)
        return fileName

    def uploadLog(self, fileName, uploadName=None):
        """
        """
        if not uploadName:
            uploadName = os.path.split(fileName)[1]
        params = [('name', uploadName)]

        resString = self["requests"].uploadFile(fileName=fileName, fieldName='inputfile',
                                                url=self['endpoint'] + 'logfile',
                                                params=params, verb='PUT')

        return json.loads(resString)['result'][0]

    def removeFile(self, haskey):
        result=self['requests'].makeRequest(uri = 'info', data = {'subresource':'fileremove', 'hashkey': haskey})
        return json.loads(result[0])['result'][0]

    def download(self, hashkey, output, username=None):
        """
        Download tarfile with the provided hashkey.
        """
        url = self['endpoint'] + 'file?hashkey=%s' % hashkey
        if username:
            url = url + '&username=%s' % username

        self['logger'].info('Fetching URL %s' % url)
        fileName, dummyHeader = self['requests'].downloadFile(output, str(url)) # unicode broke pycurl.setopt
        self['logger'].debug('Wrote %s' % fileName)
        return fileName

    def upload(self, fileName, excludeList = None):
        """
        Upload the tarfile fileName to the user file cache. Returns the hash of the content of the file
        which can be used to retrieve the file later on.
        """
        if excludeList==None: #pylint says [] is a dangerous default value
            excludeList = []

        #The parameter newchecksum tells the crabcache to use the new algorithm. It's there
        #to guarantee backward compatibility.
        params = [('hashkey', calculateChecksum(fileName, excludeList)), ('newchecksum', '2')]

        resString = self["requests"].uploadFile(fileName=fileName, fieldName='inputfile',
                                                url=self['endpoint'] + 'file',
                                                params=params, verb='PUT')

        return json.loads(resString)['result'][0]




#!/usr/bin/env python
"""
_SetCache_

Oracle implementation of Jobs.SetCache
"""

__all__ = []



from WMCore.WMBS.MySQL.Jobs.SetCache import SetCache as MySQLSetCache

class SetCache(MySQLSetCache):
    """
    Identical to MySQL version for now

    """



#!/usr/bin/env python
"""
_GetStateID_

Oracle implementation of Jobs.GetStateID

Created on Sep 13, 2012

@author: dballest
"""

from WMCore.WMBS.MySQL.Jobs.GetStateID import GetStateID as MySQLGetStateID

class GetStateID(MySQLGetStateID):
    pass



#!/usr/bin/env python
#pylint: disable=E1101,E1103,C0103,R0902

"""
Makes jobs in the proper state.
"""
from __future__ import print_function




import logging

# harness class that encapsulates the basic component logic.
from WMCore.Agent.Harness import Harness
from WMCore.WMFactory import WMFactory



class JobMaker(Harness):

    def __init__(self, config):
        # call the base class
        Harness.__init__(self, config)
        print("JobMaker.__init__")

    def preInitialization(self):
        print("JobMaker.preInitialization")

        # use a factory to dynamically load handlers.
        factory = WMFactory('generic')
        self.messages['MakeJob'] = \
            factory.loadObject('WMCore.WMSpec.Makers.Handlers.MakeJob', self)



#!/usr/bin/env python
"""
_ListRunningJobs_

Oracle implementation of Monitoring.ListRunningJobs
"""




from WMCore.WMBS.MySQL.Monitoring.ListRunningJobs import ListRunningJobs \
    as ListRunningJobsMySQL

class ListRunningJobs(ListRunningJobsMySQL):
    pass



#!/usr/bin/env python
"""
_GetNumberOfJobsPerSite_

MySQL implementation of Jobs.GetNumberOfJobsPerSite
"""

__all__ = []



import logging

from WMCore.Database.DBFormatter import DBFormatter

class GetNumberOfJobsPerSite(DBFormatter):
    """
    _GetLocation_

    Retrieve all files that are associated with the given job from the
    database.
    """
    sql = """SELECT count(*) FROM wmbs_job
           WHERE location = (SELECT ID FROM wmbs_location WHERE site_name = :location)
           AND jobgroup IN (SELECT ID FROM wmbs_jobgroup jg WHERE jg.subscription = :subscription)
           AND state = (SELECT ID FROM wmbs_job_state js WHERE js.name = :state)
           """

    def format(self, results):
        """
        _format_

        """

        if len(results) == 0:
            return False
        else:
            return results[0].fetchall()[0]


    def buildBinds(self, location, subscription, state):
        """

        _buildBinds_

        Build a list of binds

        """

        binds = {}

        binds['location']     = location
        binds['subscription'] = subscription
        binds['state' ]       = state

        return binds


    def execute(self, location, subscription, state, conn = None, transaction = False):
        """
        _execute_

        Execute the SQL for the given job ID and then format and return
        the result.
        """



        binds = self.buildBinds(location, subscription, state)

        result = self.dbi.processData(self.sql, binds, conn = conn, transaction = transaction)


        return self.format(result)



#!/usr/bin/env python

"""
Oracle implementation of AddChecksum
"""





from WMCore.WMBS.MySQL.Files.AddChecksum import AddChecksum as MySQLAddChecksum

class AddChecksum(MySQLAddChecksum):
    """
    Identical to MySQL Version

    """



#!/usr/bin/env python
"""
__init__

Module for generating lifecycle workflow component stubs,
so developers can easily migrate their handlers without
worrying to much about the structure and framework.


"""



#!/usr/bin/env python
"""
_BasicRun_

Basic run and lumi sequence generator

No state storage.

"""
from WMCore.JobSplitting.Generators.GeneratorInterface import GeneratorInterface

class BasicRun(GeneratorInterface):

    def __init__(self, **options):
        GeneratorInterface.__init__(self, **options)

        self.initialRun    = int(options.get("initial_run",   1))
        self.runIncrement  = int(options.get("run_increment", 1))
        self.initialLumi   = int(options.get("initial_lumi",  0))
        self.lumiIncrement = int(options.get("lumi_increment",1))
        self.lumiPerRun    = int(options.get("lumi_per_run", 10))

        self.currentRun = self.initialRun
        self.currentLumi = self.initialLumi
        self.lumiCount = 0


    def incrementRun(self):
        self.currentRun += self.runIncrement


    def incrementLumi(self):
        if self.lumiCount == self.lumiPerRun:
            self.currentLumi = 0
            self.lumiCount = 0
            self.incrementRun()
        else:
            self.currentLumi += self.lumiIncrement
            self.lumiCount += 1
        return

    def __call__(self, wmbsJob):

        wmbsJob['mask']['FirstRun'] = self.currentRun
        wmbsJob['mask']['FirstLumi'] = self.currentLumi

        self.incrementLumi()



#!/usr/bin/env python
"""
_LoadByID_

Oracle implementation for loading a job by scheduler status
"""


from WMCore.BossAir.MySQL.LoadByID import LoadByID as MySQLLoadByID

class LoadByID(MySQLLoadByID):
    """
    _LoadByID_

    Load all jobs in full by ID
    """



#!/usr/bin/env python
"""
_MarkOpen_

MySQL implementation of Fileset.MarkOpen
"""

__all__ = []

import time

from WMCore.Database.DBFormatter import DBFormatter

class MarkOpen(DBFormatter):
    sql = "UPDATE wmbs_fileset SET open = :p_1, last_update = :time WHERE name = :p_2"

    def execute(self, fileset = None, isOpen = True, conn = None,
                transaction = False):
        bindVars = {"p_1": int(isOpen), "p_2": fileset, "time": int(time.time())}
        self.dbi.processData(self.sql, bindVars, conn = conn,
                             transaction = transaction)
        return



#!/usr/bin/env python
"""
_RandomSeeder_

Simple Random seed generator

"""
import random

from WMCore.JobSplitting.Generators.GeneratorInterface import GeneratorInterface

class RandomSeeder(GeneratorInterface):
    """
    Seeder that generates a set of random numbers for a job
    and inserts them into the job baggage

    """
    def __init__(self, **options):
        self.options  = options
        self._MAXINT  = options.get('MAXINT', 900000000)

    def __call__(self, wmbsJob):
        baggage = wmbsJob.getBaggage()
        for x in self.options:
            wmbsJob.addBaggageParameter("RandomSeeder.%s" %(x), random.randint(1, self._MAXINT))



#!/usr/bin/env python
#pylint: disable=E1101,E1103,C0103,R0902
"""
DefaultConfig.py

Sample configuration for generating workflow.

"""





import os
import pickle

from WMCore.Agent.Configuration import Configuration
config = Configuration()
config.section_('General')

# output directory of generated files.
config.General.baseDir = '/tmp/PRODAGENT'
config.General.srcDir = os.path.join(config.General.baseDir, 'src/python/PA/Component')
config.General.testDir =  os.path.join(config.General.baseDir, 'test/python/PA_t/Component_t')
# module prefix for python files.
config.General.pythonPrefix = 'PA.Component'
config.General.pythonTestPrefix = 'PA_t.Component_t'
config.General.handlers = []
config.General.synchronizers = []
config.General.plugins = []

# synchronizer is for trigger module.
synchronizer = {'ID' : 'JobPostProcess', \
                'action' : 'PA.Core.Trigger.PrepareCleanup'}
config.General.synchronizers.append(pickle.dumps(synchronizer))

# A handler is a piece of code that takes as input a message (and its payload)
# and performs certain actions. For example: a submit job handler takes as
# input messages of type SubmitJob with payload a job specification, and
# submits the #job to a particular site. Handlers are grouped into components.
# For example a submit failure handler and process failure handler can be
# grouped into an autonomous

# a handler spec has (at most) 5 attributes:
# -messageIn. (what message the component subscribes to should this handler
# handle.
# -messageOut (optional). What messages should this message publish.
# Messages separated
# by a comma mean that all these messages need to be published. Messages
# separated by a | means either or.
# -component. To which cmponent is this handler associated.
# -threading (optional). Are the messages for this handler dispatched to threads
# or handled sequential.
# createSynchronizer (optional). When this handler is finished handling a
# message should it create a trigger (so other components can set a flag).
# -synchronize. After handling the message this handler needs to set a flag
# in a particular synchronizer (aka trigger).
# configurable. Should the handlers that are associated to a message be
# configurable in the associated config file of the component.

handler = {'messageIn'   : 'CreateJob', \
           'messageOut'  : 'SubmitJob|JobCreateFailed', \
           'component'   : 'JobCreator'}
config.General.handlers.append(pickle.dumps(handler))

handler = {'messageIn'   : 'ReCreateJob', \
           'messageOut'  : 'SubmitJob|JobCreateFailed', \
           'component'   : 'JobCreator'}
config.General.handlers.append(pickle.dumps(handler))

handler = {'messageIn'   : 'SubmitJob', \
           'messageOut'  : 'TrackJob|JobSubmitFailed', \
           'component'   : 'JobSubmitter', \
           'threading'   : 'yes', \
           'createSynchronizer' : 'JobPostProcess'}
config.General.handlers.append(pickle.dumps(handler))

handler = {'messageIn'   : 'ReSubmitJob', \
           'messageOut'  : 'TrackJob|JobSubmitFailed', \
           'component'   : 'JobSubmitter', \
           'createSynchronizer' : 'JobPostProcess'}
config.General.handlers.append(pickle.dumps(handler))

handler = {'messageIn'   : 'TrackJob', \
           'messageOut'  : 'JobProcessSuccess|JobProcessFailed', \
           'component'   : 'JobTracker', \
           'threading'   : 'yes'}
config.General.handlers.append(pickle.dumps(handler))

handler = {'messageIn'   : 'JobProcessFailed', \
           'configurable': 'yes', \
           'messageOut'  : 'ResubmitJob|JobFailed', \
           'component'   : 'ErrorHandler'}
config.General.handlers.append(pickle.dumps(handler))

handler = {'messageIn'   : 'JobSubmitFailed', \
           'configurable': 'yes', \
           'messageOut'  : 'ResubmitJob|JobFailed', \
           'component'   : 'ErrorHandler'}
config.General.handlers.append(pickle.dumps(handler))

handler = {'messageIn'   : 'JobCreateFailed', \
           'configurable': 'yes', \
           'messageOut'  : 'ResubmitJob|JobFailed', \
           'component'   : 'ErrorHandler'}
config.General.handlers.append(pickle.dumps(handler))

handler = {'messageIn'   : 'CleanJob', \
           'messageOut'  : '', \
           'component'   : 'JobCleanup'}
config.General.handlers.append(pickle.dumps(handler))

handler = {'messageIn'   : 'JobFailed', \
           'messageOut'  : '', \
           'component'   : 'JobCleanup'}
config.General.handlers.append(pickle.dumps(handler))

handler = {'messageIn'   : 'JobSuccess', \
           'messageOut'  : '', \
           'component'   : 'JobCleanup'}
config.General.handlers.append(pickle.dumps(handler))

handler = {'messageIn'   : 'JobProcessSuccess', \
           'messageOut'  : '', \
           'component'   : 'DBS', \
           'synchronize' : 'JobPostProcess'}
config.General.handlers.append(pickle.dumps(handler))

handler = {'messageIn'   : 'JobProcessSucess', \
           'messageOut'  : 'AccountData', \
           'component'   : 'MergeSensor', \
           'synchronize' : 'JobPostProcess', \
           'threading'   : 'yes'}
config.General.handlers.append(pickle.dumps(handler))

handler = {'messageIn'   : 'JobProcessSuccess', \
           'messageOut'  : '', \
           'component'   : 'Phedex', \
           'synchronize' : 'JobPostProcess'}
config.General.handlers.append(pickle.dumps(handler))


plugin = {'component'  : 'JobSubmitter', \
          'plugins'    : 'PyCondor', \
          'handler'    : 'SubmitJob'}
config.General.plugins.append(pickle.dumps(plugin))



#!/usr/bin/env python
#-*- coding: utf-8 -*-
#pylint: disable=
"""
File       : Signals.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: This module provide set of hooks to running program
The dumpthreads callback function will be called every time when
running program will receive SIGUSR1 signal. E.g., in unix shell
just do
shell# kill -s SIGUSR1 <pid>
"""
# future
from __future__ import print_function, division
from future.utils import viewitems

# system modules
import sys
import signal
import threading
import traceback

def dumpthreads(isignal, iframe):
    """
    Dump context of all threads upon given signal
    http://stackoverflow.com/questions/132058/showing-the-stack-trace-from-a-running-python-application
    """
    print("DAS stack, signal=%s, frame=%s" % (isignal, iframe))
    id2name = dict([(th.ident, th.name) for th in threading.enumerate()])
    code = []
    for tid, stack in viewitems(sys._current_frames()):
        code.append("\n# Thread: %s(%d)" % (id2name.get(tid,""), tid))
        for filename, lineno, name, line in traceback.extract_stack(stack):
            code.append('File: "%s", line %d, in %s' % (filename, lineno, name))
            if  line:
                code.append("  %s" % (line.strip()))
    print("\n".join(code))

signal.signal(signal.SIGUSR1, dumpthreads)



#!/usr/bin/env python
"""
_StepFactory_

Factory implementation to retrieve Step Template instances

"""


from WMCore.WMFactory import WMFactory
from WMCore.WMException import WMException

class StepFactoryException(WMException):
    """
    _StepFactortyException_

    Exception for missing objects or problems

    """
    pass



class TemplateFactory(WMFactory):
    """
    _TemplateFactory_

    Instantiate a WMFactory instance with the appropriate namespace

    """
    def __init__(self):
        WMFactory.__init__(self, self.__class__.__name__,
                           "WMCore.WMSpec.Steps.Templates")

class BuilderFactory(WMFactory):
    """
    _BuilderFactory_

    Instantiate a WMFactory instance with the appropriate namespace

    """
    def __init__(self):
        WMFactory.__init__(self, self.__class__.__name__,
                           "WMCore.WMSpec.Steps.Builders")

class ExecutorFactory(WMFactory):
    """
    _ExecutorFactory_

    Instantiate a WMFactory instance with the appropriate namespace

    """
    def __init__(self):
        WMFactory.__init__(self, self.__class__.__name__,
                           "WMCore.WMSpec.Steps.Executors")
class EmulatorFactory(WMFactory):
    """
    _EmulatorFactory_

    Instantiate a WMFactory instance with the appropriate namespace

    """
    def __init__(self):
        WMFactory.__init__(self, self.__class__.__name__,
                           "WMCore.WMSpec.Steps.Emulators")

class FetcherFactory(WMFactory):
    """
    _FetcherFactory_

    Instantiate a WMFactory instance for the Fetchers package

    """
    def __init__(self):
        WMFactory.__init__(self, self.__class__.__name__,
                           "WMCore.WMSpec.Steps.Fetchers")

class DiagnosticFactory(WMFactory):
    """
    _DiagnosticFactory_

    Instantiate a WMFactory instance for the Diagnostics package

    """
    def __init__(self):
        WMFactory.__init__(self, self.__class__.__name__,
                           "WMCore.WMSpec.Steps.Diagnostics")



_TemplateFactory = TemplateFactory()
_BuilderFactory  = BuilderFactory()
_ExecutorFactory = ExecutorFactory()
_EmulatorFactory = EmulatorFactory()
_FetcherFactory  = FetcherFactory()
_DiagnosticFactory = DiagnosticFactory()

def getStepTemplate(stepType):
    """
    _getStepTemplate_

    factory method to return a step template instance based on the
    name of the step

    """
#    try:
    return _TemplateFactory.loadObject(stepType)
#    except WMException, wmEx:
#        msg = "TemplateFactory Unable to load Object: %s" % stepType
#        raise StepFactoryException(msg)
#    except Exception, ex:
#        msg = "Error creating object %s in TemplateFactory:\n" % stepType
#        msg += str(ex)
#        raise StepFactoryException(msg)

def getStepTypeHelper(stepReference):
    """
    _getStepTypeHelper_

    Given a step instance, get its type, use that to get a template
    from the factory and use the template to instantiate the type
    specific helper

    """
    stepType = getattr(stepReference, "stepType", None)
    if stepType == None:
        msg = "Unable to find stepType attribute for step reference passed\n"
        msg += "to getStepTypeHelper method"
        raise StepFactoryException(msg)
    template = getStepTemplate(stepType)
    helper = template.helper(stepReference)
    return helper


def getStepBuilder(stepType):
    """
    _getStepBuilder_

    """
    #try:
    return _BuilderFactory.loadObject(stepType)
#    except WMException, wmEx:
#        msg = "BuilderFactory Unable to load Object: %s" % stepType
#        raise StepFactoryException(msg)
#    except Exception, ex:
#        msg = "Error creating object %s in BuilderFactory:\n" % stepType
#        msg += str(ex)
#        raise StepFactoryException(msg)

def getStepExecutor(stepType):
    """
    _getStepExecutor_

    Get an Executor for the given step type

    """
#    try:
    executor =_ExecutorFactory.loadObject(stepType)
#    except WMException, wmEx:
#        msg = "ExecutorFactory Unable to load Object: %s" % stepType
#        raise StepFactoryException(msg)
#    except Exception, ex:
#        msg = "Error creating object %s in ExecutorFactory:\n" % stepType
#        msg += str(ex)
#        raise StepFactoryException(msg)

    executor.diagnostic = getDiagnostic(stepType)
    return executor

def getStepEmulator(stepEmuName):
    """
    _getStepEmulator_

    Get an instance of a given emulator, Note that this takes the name
    of the emulator rather than the step Type since there are multiple
    ways to emulate a given step

    """
    #try:
    return _EmulatorFactory.loadObject(stepEmuName)
    #except WMException, wmEx:
    #    msg = "EmulatorFactory Unable to load Object: %s" % stepEmuName
    #    raise StepFactoryException(msg)
    #except Exception, ex:
    #    msg = "Error creating object %s in EmulatorFactory:\n" % stepEmuName
    #    msg += str(ex)
    #    raise StepFactoryException(msg)

def getFetcher(fetcherName):
    """
    _getFetcher_

    Get an instance of the named Fetcher implementation

    """
    return _FetcherFactory.loadObject(fetcherName)

def getDiagnostic(stepType):
    """
    _getDiagnostic_

    Get the appropriate Diagnostic for the stepType, falling back
    on a Generic Diagnostic if there isnt a specific impl

    """
    try:
        return _DiagnosticFactory.loadObject(stepType)
    except WMException as wmEx:
        return _DiagnosticFactory.loadObject("Generic")
#    except Exception, ex:
#        msg = "Error creating object %s in DiagnosticFactory:\n" % stepType
#        msg += str(ex)
#        raise StepFactoryException(msg)



#!/usr/bin/env python
"""
_StageInMgr_

Util class to provide stage in functionality as an interface object.

Based on StageOutMgr class

"""
from __future__ import print_function

from builtins import object
from future.utils import viewitems

import os

from WMCore.Storage.SiteLocalConfig import loadSiteLocalConfig

#do we want seperate exceptions - for the moment no
from WMCore.Storage.StageOutError import StageOutFailure
from WMCore.Storage.StageOutError import StageOutInitError
from WMCore.Storage.Registry import retrieveStageOutImpl


class StageInSuccess(Exception):
    """
    _StageOutSuccess_

    Exception used to escape stage out loop when stage out is successful
    """
    pass



class StageInMgr(object):
    """
    _StageInMgr_

    Object that can be used to stage out a set of files
    using TFC or an override.

    """
    def __init__(self, **overrideParams):
        self.override = False
        self.overrideConf = overrideParams
        if overrideParams != {}:
            self.override = True

        self.fallbacks = []
        #  //
        # // Try an get the TFC for the site
        #//
        self.tfc = None



        self.numberOfRetries = 3
        self.retryPauseTime = 600

        #  //
        # // If override isnt None, we dont need SiteCfg, if it is
        #//  then we need siteCfg otherwise we are dead.

        if self.override == False:
            self.siteCfg = loadSiteLocalConfig()

        if self.override:
            self.initialiseOverride()
        else:
            self.initialiseSiteConf()


    def initialiseSiteConf(self):
        """
        _initialiseSiteConf_

        Extract required information from site conf and TFC

        """

        implName = self.siteCfg.localStageOut.get("command", None)
        if implName == None:
            msg = "Unable to retrieve local stage out command\n"
            msg += "From site config file.\n"
            msg += "Unable to perform StageOut operation"
            raise StageOutInitError( msg)
        msg = "Local Stage Out Implementation to be used is:"
        msg += "%s\n" % implName

        pnn = self.siteCfg.localStageOut.get("phedex-node", None)
        if pnn == None:
            msg = "Unable to retrieve local stage out phedex-node\n"
            msg += "From site config file.\n"
            msg += "Unable to perform StageOut operation"
            raise StageOutInitError( msg)
        msg += "Local Stage Out PNN to be used is %s\n" % pnn
        catalog = self.siteCfg.localStageOut.get("catalog", None)
        if catalog == None:
            msg = "Unable to retrieve local stage out catalog\n"
            msg += "From site config file.\n"
            msg += "Unable to perform StageOut operation"
            raise StageOutInitError( msg)
        msg += "Local Stage Out Catalog to be used is %s\n" % catalog

        try:
            self.tfc = self.siteCfg.trivialFileCatalog()
            msg += "Trivial File Catalog has been loaded:\n"
            msg += str(self.tfc)
        except Exception as ex:
            msg = "Unable to load Trivial File Catalog:\n"
            msg += "Local stage out will not be attempted\n"
            msg += str(ex)
            raise StageOutInitError( msg )

        print(msg)
        return


    def initialiseOverride(self):
        """
        _initialiseOverride_

        Extract and verify that the Override parameters are all present

        """
        overrideConf = self.overrideConf
        overrideParams = {
            "command" : None,
            "option" : None,
            "phedex-node" : None,
            "lfn-prefix" : None,
            }

        try:
            overrideParams['command'] = overrideConf['command']
            overrideParams['phedex-node'] = overrideConf['phedex-node']
            overrideParams['lfn-prefix'] = overrideConf['lfn-prefix']
        except Exception as ex:
            msg = "Unable to extract Override parameters from config:\n"
            msg += str(overrideConf)
            raise StageOutInitError(msg)
        if 'option' in overrideConf:
            if len(overrideConf['option']) > 0:
                overrideParams['option'] = overrideConf['option']
            else:
                overrideParams['option'] = ""

        msg = "=======StageIn Override Initialised:================\n"
        for key, val in viewitems(overrideParams):
            msg += " %s : %s\n" % (key, val)
        msg += "=====================================================\n"
        print(msg)
        self.fallbacks = []
        self.fallbacks.append(overrideParams)
        return


    def __call__(self, **fileToStage):
        """
        _operator()_

        Use call to invoke transfers

        """


        try:
            print("==>Working on file: %s" % fileToStage['LFN'])

            lfn = fileToStage['LFN']

            #  //
            # // No override => use local-stage-out from site conf
            #//  invoke for all files and check failures/successes
            if not self.override:
                print("===> Attempting Local Stage In.")
                try:
                    pfn = self.localStageIn(lfn)
                    fileToStage['PFN'] = pfn
                    raise StageInSuccess
                except StageOutFailure as ex:
                    msg = "===> Local Stage Out Failure for file:\n"
                    msg += "======>  %s\n" % fileToStage['LFN']
                    msg += str(ex)
                    print(msg)
            #  //
            # // Still here => override start using the fallback stage outs
            #//  If override is set, then that will be the only fallback available
            print("===> Attempting %s Override Stage Outs" % len(self.fallbacks))
            for fallback in self.fallbacks:
                try:
                    pfn = self.localStageIn(lfn, fallback)
                    fileToStage['PFN'] = pfn
                    raise StageInSuccess
                except StageOutFailure as ex:
                    continue

        except StageInSuccess:
            msg = "===> Stage In Successful:\n"
            msg += "====> LFN: %s\n" % fileToStage['LFN']
            msg += "====> PFN: %s\n" % fileToStage['PFN']
            print(msg)
            return fileToStage
        msg = "Unable to stage out file:\n"
        msg += fileToStage['LFN']
        raise StageOutFailure(msg, **fileToStage)


    def localStageIn(self, lfn, override = None):
        """
        _localStageOut_

        Given the lfn and local stage out params, invoke the local stage in
        i.e. stage in lfn to pfn

        if override is used the follwoing params should be defined:
        command - the stage out impl plugin name to be used
        option - the option values to be passed to that command (None is allowed)
        lfn-prefix - the LFN prefix to generate the PFN
        phedex-node - the Name of the PNN to which the file is being xferred

        """
        localPfn = os.path.join(os.getcwd(), os.path.basename(lfn))

        if override:
            pnn = override['phedex-node']
            command = override['command']
            options = override['option']
            pfn = "%s%s" % (override['lfn-prefix'], lfn)
            protocol = command
        else:
            pnn = self.siteCfg.localStageOut['phedex-node']
            command = self.siteCfg.localStageOut['command']
            options = self.siteCfg.localStageOut.get('option', None)
            pfn = self.searchTFC(lfn)
            protocol = self.tfc.preferredProtocol

        if pfn == None:
            msg = "Unable to match lfn to pfn: \n  %s" % lfn
            raise StageOutFailure(msg, LFN = lfn, TFC = str(self.tfc))

        try:
            impl = retrieveStageOutImpl(command, stagein=True)
        except Exception as ex:
            msg = "Unable to retrieve impl for local stage in:\n"
            msg += "Error retrieving StageOutImpl for command named: %s\n" % (
                command,)
            raise StageOutFailure(msg, Command = command,
                                  LFN = lfn, ExceptionDetail = str(ex))
        impl.numRetries = self.numberOfRetries
        impl.retryPause = self.retryPauseTime

        try:
            impl(protocol, pfn, localPfn, options)
        except Exception as ex:
            msg = "Failure for local stage in:\n"
            msg += str(ex)
            raise StageOutFailure(msg, Command = command, Protocol = protocol,
                                  LFN = lfn, InputPFN = localPfn,
                                  TargetPFN = pfn)

        return localPfn


    def searchTFC(self, lfn):
        """
        _searchTFC_

        Search the Trivial File Catalog for the lfn provided,
        if a match is made, return the matched PFN

        """
        if self.tfc == None:
            msg = "Trivial File Catalog not available to match LFN:\n"
            msg += lfn
            print(msg)
            return None
        if self.tfc.preferredProtocol == None:
            msg = "Trivial File Catalog does not have a preferred protocol\n"
            msg += "which prevents local stage out for:\n"
            msg += lfn
            print(msg)
            return None

        pfn = self.tfc.matchLFN(self.tfc.preferredProtocol, lfn)
        if pfn == None:
            msg = "Unable to map LFN to PFN:\n"
            msg += "LFN: %s\n" % lfn
            return None

        msg = "LFN to PFN match made:\n"
        msg += "LFN: %s\nPFN: %s\n" % (lfn, pfn)
        print(msg)
        return pfn



#/usr/bin/env python
"""
_Destroy_

Oracle implementation of ResourceControl.Destroy.
"""




from WMCore.ResourceControl.MySQL.Destroy import Destroy as MySQLDestroy

class Destroy(MySQLDestroy):
    pass



#!/usr/bin/env python
"""
_LoadComplete_

Oracle implementation for loading bl_runjob records for complete jobs
"""


from WMCore.BossAir.MySQL.LoadComplete import LoadComplete as MySQLLoadComplete

class LoadComplete(MySQLLoadComplete):
    """
    _LoadComplete_

    """



#!/usr/bin/env python
"""
_XRDCPImpl_

Implementation of StageOutImpl interface for RFIO in Castor-2

"""
import os
import logging

from WMCore.Storage.StageOutImplV2 import StageOutImplV2
from WMCore.WMBase import getWMBASE
from subprocess import Popen, PIPE
from WMCore.Storage.Execute import runCommand
from WMCore.Storage.StageOutError import StageOutError, StageOutFailure


class XRDCPImpl(StageOutImplV2):
    """
    _XRDCPImpl_

    Implement interface for rfcp command

    """

    def doWrapped(self, commandArgs):
        wrapperPath = os.path.join(getWMBASE(),'src','python','WMCore','Storage','Plugins','XRDCP','wrapenv.sh')
        commandArgs.insert(0,wrapperPath)
        return runCommand(commandArgs)

    def doTransfer(self, sourcePFN, targetPFN, stageOut, pnn, command, options, protocol, checksum ):
        """
            performs a transfer. stageOut tells you which way to go. returns the new pfn or
            raises on failure. StageOutError (and inherited exceptions) are for expected errors
            such as temporary connection failures. Anything else will be handled as an unexpected
            error and skip retrying with this plugin
        """
        original_size = os.stat(sourcePFN)[6]
        logging.info("Local File Size is: %s" % original_size)
        self.doWrapped(self.generateCommandFromPreAndPostParts(['xrdcp','-s3'],
                                                               [sourcePFN,'root://lxgate39.cern.ch/%s'% targetPFN],
                                                               options))

        p1 = Popen(["rfstat", '-recursion_depth=0','-retry_num=0', targetPFN], stdout=PIPE)
        p2 = Popen(["grep", 'Size'], stdout=PIPE, stdin=p1.stdout)
        p3 = Popen(["cut", '-f2', '-d:'], stdout=PIPE, stdin=p2.stdout)
        remoteSize = p3.communicate()[0]

        if int(original_size) != int(remoteSize):
            try:
                self.doDelete(targetPFN, None, None, None, None)
            except:
                pass
            raise StageOutFailure("File sizes don't match")

    def doDelete(self, pfnToRemove, pnn, command, options, protocol  ):
        """
        _removeFile_

        CleanUp pfn provided: specific for Castor-1

        """
        self.runCommandWarnOnNonZero(['stager_rm','-M',pfnToRemove])
        self.runCommandWarnOnNonZero(['nsrm',pfnToRemove])



#!/usr/bin/env python
"""
_AddLocation_

MySQL implementation of DBSBufferFiles.AddLocation
"""

from builtins import str, bytes
from WMCore.Database.DBFormatter import DBFormatter

class AddLocation(DBFormatter):
    sql = """INSERT IGNORE INTO dbsbuffer_location (pnn)
               VALUES (:location)"""

    def execute(self, siteName, conn = None, transaction = False):
        binds = []
        if isinstance(siteName, (str, bytes)):
            siteName = [siteName]

        for site in siteName:
            binds.append({"location": site})

        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = True)
        return
